import itertools
from typing import List, Any
import torch
import pickle
from .stdvae import StandardVAE
from .mulvae import MultipleVAE
from .utils.models.learning_utils import make_train_step, early_stopping
import os
import pandas as pd

# Global variables which translate the names into the corresponding functions and objects
activation_fn_dict = {'tanh': torch.nn.Tanh, 'relu': torch.nn.ReLU, 'sigmoid': torch.nn.Sigmoid,
                      'leaky relu': torch.nn.LeakyReLU}
optimizers_dict = {'adam': torch.optim.Adam, 'Adadelta': torch.optim.Adadelta, 'Adagrad': torch.optim.Adagrad}


class Data_inegrator:
    def __init__(self, parse_args):
        """
        The idea of this classs, is to transform the inputs inserted by the user into a data integration tool. The
        Data_inegrator will iterate over all different configurations and exploit the best one to project the data
        into its latent representation.
        :param parse_args: which can be achieved by using the parser module in the parser package.
        """
        with open(parse_args.data_path, 'rb') as data_file:
            self.patients_dataset = pickle.load(data_file)
        self.results_path = parse_args.results_path
        # find the patient which consists of both saliva and stool, and the patient without one of them.
        indexes_of_patients_with_all_fields, indexes_of_patients_with_field0_only, indexes_of_patients_with_field1_only = self.patients_dataset.separate_to_groups()
        self.xy_dataset = torch.utils.data.Subset(self.patients_dataset, indexes_of_patients_with_all_fields)
        self.x_dataset = torch.utils.data.Subset(self.patients_dataset, indexes_of_patients_with_field0_only)
        self.y_dataset = torch.utils.data.Subset(self.patients_dataset, indexes_of_patients_with_field1_only)

        self.x_architecture = parse_args.x_architecture
        self.y_architecture = parse_args.y_architecture
        self.xy_architecture = parse_args.xy_architecture

        # train and validate only according to the patients with both fields.

        train_size = int(parse_args.train_size * len(self.xy_dataset))
        validation_size = len(self.xy_dataset) - train_size

        # split the data into train and validation
        self.xy_train_dataset, self.xy_validation_dataset = torch.utils.data.random_split(self.xy_dataset,
                                                                                          [train_size, validation_size])
        # Creating the dataloaders for the xy_dataset.

        self.xy_train_dataloader = torch.utils.data.DataLoader(self.xy_train_dataset, shuffle=True,
                                                               batch_size=parse_args.batch_size)

        self.xy_validation_dataloader = torch.utils.data.DataLoader(self.xy_validation_dataset,
                                                                    batch_size=len(self.xy_validation_dataset),
                                                                    shuffle=True)

        self.learning_rate_list = parse_args.learning_rate
        self.activation_fn = activation_fn_dict[parse_args.activation_fn]()
        self.optimizer = optimizers_dict[parse_args.optimizer]
        self.latent_layer_size_list = parse_args.latent_representation
        self.klb_coefficient_list = parse_args.klb_coefficient
        self.patience = parse_args.patience

        self.configuration_list = self._create_all_configurations()

    def _create_all_configurations(self) -> List[dict]:
        """
        Create all possible combinations of (learning_rate.latent_layer_size,klb_coefficient) that can be extracted
        from the user's input
        :return: A list of configurations, where each one is represented by a dictionary.
        """
        parameters_list = itertools.product(self.learning_rate_list, self.latent_layer_size_list,
                                            self.klb_coefficient_list)
        configuration_list = [
            {'learning_rate': learning_rate, 'latent_layer_size': latent_layer_size, 'klb_coefficient': klb_coefficient}
            for learning_rate, latent_layer_size, klb_coefficient in parameters_list]
        return configuration_list

    def find_best_configuration(self):
        best_results_of_each_configuration = []
        best_model_loss = 10 ** 10
        for configuration in self.configuration_list:
            learning_rate = configuration['learning_rate']
            latent_layer_size = configuration['latent_layer_size']
            klb_coefficient = configuration['klb_coefficient']

            print(
                'A model with LR : {lr} and latent_size : {latent_size} and klb coefficient : {klb}  Is now runing'.format(
                    lr=learning_rate, latent_size=latent_layer_size,
                    klb=klb_coefficient))

            # Construct the models according to the current configuration
            xy_vae = StandardVAE(self.xy_architecture + [latent_layer_size], self.activation_fn)
            x_vae = StandardVAE(self.x_architecture + [latent_layer_size], self.activation_fn)
            y_vae = StandardVAE(self.y_architecture + [latent_layer_size], self.activation_fn)
            # Create the full VAE based on the standardVAE's above.
            full_vae = MultipleVAE(xy_vae, x_vae, y_vae)
            optimizer = self.optimizer(full_vae.parameters(), lr=learning_rate)
            train_step_function = make_train_step(full_vae, optimizer)

            # Train the model
            average_validation_sample_loss_per_epoch = []
            epoch = 0

            stopping_epoch = 0

            # Train full model.

            stop = False

            # Use early stopping.

            while not stop:

                epoch += 1
                # Do an epoch on the training set.

                for patient_train_batch in self.xy_train_dataloader:
                    field0_batch, field1_batch = patient_train_batch['FIELD0'], patient_train_batch['FIELD1']
                    # If for some reason the batch has mor dimensions than expected, squeeze it.
                    if len(field0_batch.shape) > 2:
                        field0_batch, field1_batch = patient_train_batch['FIELD0'].squeeze(), patient_train_batch[
                            'FIELD1'].squeeze()

                    train_step_function(field0_batch, field1_batch)
                # Do validation
                total_validation_loss_per_epoch = 0
                with torch.no_grad():
                    full_vae.eval()
                    for validation_patient_batch in self.xy_validation_dataloader:
                        field0_batch, field1_batch = validation_patient_batch['FIELD0'].squeeze(), \
                                                     validation_patient_batch[
                                                         'FIELD1'].squeeze()
                        forward_dict = full_vae(field0_batch, field1_batch)
                        # Computes reconstruction loss and according to it decide whether to stop.
                        loss_dict = full_vae.loss_function(forward_dict)
                        total_validation_loss_per_epoch += float(loss_dict['xvae_loss']['Reconstruction_Loss'] + \
                                                                 loss_dict['xyvae_loss']['Reconstruction_Loss'] + \
                                                                 loss_dict['yvae_loss']['Reconstruction_Loss'])
                    # Compute the average validation sample loss.
                    average_validation_sample_loss = total_validation_loss_per_epoch / len(self.xy_validation_dataset)
                    # If the loss achieved on the validation is the smallest until so far, save the model.
                    if average_validation_sample_loss < best_model_loss:
                        torch.save(full_vae, os.path.join(self.results_path, 'best_model.pt'))
                        best_model_loss = average_validation_sample_loss
                    # keep tracking of the validation loss in every epoch.
                    average_validation_sample_loss_per_epoch.append(average_validation_sample_loss)
                    stop = early_stopping(average_validation_sample_loss_per_epoch, patience=self.patience,
                                          ascending=False)
            configuration_best_result = min(average_validation_sample_loss_per_epoch)
            print('The model best loss achieved on the validation set is : {}  '.format(configuration_best_result))
            best_results_of_each_configuration.append(configuration_best_result)

        configuration_results_df = pd.DataFrame(self.configuration_list)
        configuration_results_df = configuration_results_df.assign(best_loss=best_results_of_each_configuration)
        configuration_results_df.to_csv(os.path.join(self.results_path, 'models_results.csv'))

    def project_all_data(self):
        latent_representation_list = []
        patient_id_list = []
        self.patients_dataset.dict_retrieval_flag = 0
        full_vae = torch.load(os.path.join(self.results_path, 'best_model.pt'))

        with torch.no_grad():
            full_vae.eval()
            for patient in self.patients_dataset:
                if patient.get_status() == 0:
                    field0, field1 = patient.field0, patient.field1
                    try:
                        latent_representation = full_vae.xyvae(torch.cat([field0, field1], dim=0))[4]
                    except RuntimeError:
                        latent_representation = full_vae.xyvae(torch.cat([field0, field1], dim=1))[4]

                elif patient.get_status() == 1:
                    field0 = patient.field0
                    latent_representation = full_vae.xvae(field0)[4]
                else:
                    field1 = patient.field1
                    latent_representation = full_vae.yvae(field1)[4]
                latent_representation_list.append(latent_representation.squeeze().numpy())
                patient_id_list.append(patient.id)
            latent_representation_df = pd.DataFrame(data=latent_representation_list, index=patient_id_list)
            latent_representation_df.to_csv(os.path.join(self.results_path, 'latent_representation.csv'))



import pandas as pd
from torch.utils.data import Dataset
from collections.abc import Iterable

ALL_FIELDS = 0
FIELD_ZERO_ONLY = 1
FIELD_ONE_ONLY = 2


class patient_samples_single_tp():
    def __init__(self, patient_dict):
        self.patient_dict = patient_dict
        self.id = patient_dict.get('ID', 'NO_ID')
        self.time_point = patient_dict.get('TP', 'NO_TP')
        self.field0 = patient_dict.get('FIELD0', None)
        self.field1 = patient_dict.get('FIELD1', None)

    def get_status(self):
        if self.field0 is not None and self.field1 is not None:
            return ALL_FIELDS
        elif self.field0 is not None:
            return FIELD_ZERO_ONLY
        return FIELD_ONE_ONLY


class patient_dataset(Dataset):
    def __init__(self, otu_df, mapping_table, patient_id_name, time_name, sample_type_name=None, matching_fn=None,
                 transform=None):
        if matching_fn is None:
            matching_fn = lambda x, y: x == y

        self.dict_retrieval_flag = 1
        self.transform = transform
        patients_list = []
        if sample_type_name is not None:
            merged_df = pd.concat(
                [otu_df, mapping_table[patient_id_name], mapping_table[time_name], mapping_table[sample_type_name]],
                axis=1)
            unique_sample_types = list(sorted(merged_df[sample_type_name].unique()))
            for name, group in merged_df.groupby([patient_id_name, time_name]):
                patient_dict = {'ID': name[0], 'TP': [name[1]]}
                for i in range(len(group.index)):
                    current_sample = group.iloc[i]
                    current_microbiome_sample = current_sample.drop([sample_type_name, patient_id_name, time_name])
                    if unique_sample_types.index(current_sample[sample_type_name]) == 0:
                        patient_dict['FIELD0'] = current_microbiome_sample
                    else:
                        patient_dict['FIELD1'] = current_microbiome_sample
                patients_list.append(patient_samples_single_tp(patient_dict))
        else:
            all_unique_patient_ids = set.union(*[set(specific_mapping_table[specific_patient_id_name]) for
                                                 specific_mapping_table, specific_patient_id_name in
                                                 zip(mapping_table, patient_id_name)])

            for patient_id in all_unique_patient_ids:

                field_0_unique_tp, field_1_unique_tp = [
                    specific_mapping_table[specific_mapping_table[specific_patient_id_name] == patient_id][time] for
                    specific_mapping_table, specific_patient_id_name, time in
                    zip(mapping_table, patient_id_name, time_name)]
                matching_tp_list = self._find_matches(field_0_unique_tp, field_1_unique_tp, matching_fn)

                for matching_field_zero_tp, matching_field_one_tp in matching_tp_list:
                    patient_dict = {}
                    for i, (specific_otu, specific_mapping_table, patient_col_name, tp_col_name) in enumerate(
                            zip(otu_df, mapping_table, patient_id_name, time_name)):
                        if i == 0 and matching_field_zero_tp is not None:
                            patient_dict['FIELD0'] = specific_otu[
                                (specific_mapping_table[patient_col_name] == patient_id) & (specific_mapping_table[
                                                                                                tp_col_name] == matching_field_zero_tp)]

                        if i == 1 and matching_field_one_tp is not None:
                            patient_dict['FIELD1'] = specific_otu[
                                (specific_mapping_table[patient_col_name] == patient_id) & (specific_mapping_table[
                                                                                                tp_col_name] == matching_field_one_tp)]

                    if bool(patient_dict):  # check if dictionary is not empty, which means that there is no samples
                        # for the given patient_id and tp.
                        patient_dict.update({'ID': patient_id, 'TP': [matching_field_zero_tp, matching_field_one_tp]})
                    patients_list.append(patient_samples_single_tp(patient_dict))

        self.patients = patients_list

    @staticmethod
    def _find_matches(first_field_list, second_field_list, matching_fn):
        matching_couples = []
        second_field_iterator = second_field_list
        first_field_iterator = first_field_list
        for first_field in first_field_list:
            for second_field in second_field_iterator:
                if matching_fn(first_field, second_field):
                    matching_couples.append((first_field, second_field))
                    second_field_iterator = [item for item in second_field_iterator if item != second_field]
                    first_field_iterator = [item for item in first_field_iterator if item != first_field]
                    break
                else:
                    pass

        for first_field in first_field_iterator:
            matching_couples.append((first_field, None))
        for second_field in second_field_iterator:
            matching_couples.append((None, second_field))
        return matching_couples

        # don't forget to add the ones that didn't match to anybody.

    def __len__(self):
        return len(self.patient_groups)

    def __getitem__(self, idx):
        if self.transform is None:
            retrieved_patient = self.patients[idx]
        else:
            retrieved_patient = self.transform(self.patients[idx])
        return retrieved_patient.patient_dict if self.dict_retrieval_flag else retrieved_patient

    def separate_to_groups(self):
        indexes_of_patients_with_all_fields = [self.patients.index(patient) for patient in self.patients if
                                               patient.get_status() == ALL_FIELDS]
        indexes_of_patients_with_field1_only = [self.patients.index(patient) for patient in self.patients if
                                                patient.get_status() == FIELD_ZERO_ONLY]
        indexes_of_patients_with_field2_only = [self.patients.index(patient) for patient in self.patients if
                                                patient.get_status() == FIELD_ONE_ONLY]
        return indexes_of_patients_with_all_fields, indexes_of_patients_with_field1_only, indexes_of_patients_with_field2_only



import torch
import torch.utils.data
from torch import nn, optim
from torch.nn import functional as F


class MultipleVAE(nn.Module):
    def __init__(self, xyvae, xvae, yvae):
        super(MultipleVAE, self).__init__()
        """The model consists of three VAE's, one for every possible permutation of a sample (only x is given,
        only y is given, both are given) """
        self.xyvae = xyvae
        self.xvae = xvae
        self.yvae = yvae

        """Given a complete sample(x and y) forward all VAE's"""

    def forward(self, x, y):
        return {'x': self.xvae(x), 'y': self.yvae(y), 'xy': self.xyvae(torch.cat([x, y], dim=1))}

    """we're trying to minimize two losses, the  reconstruction losses of all VAE's together with the differences 
    between the latent representation of the VAE's """

    def loss_function(self, forward_dict):
        loss_dict = {'xvae_loss': self.xvae.loss_function(*(forward_dict['x'])),
                     'yvae_loss': self.xvae.loss_function(*(forward_dict['y'])),
                     'xyvae_loss': self.xvae.loss_function(*(forward_dict['xy'])),
                     'similarity_loss_x_y': F.mse_loss(forward_dict['x'][4], forward_dict['y'][4]),
                     'similarity_loss_x_xy': F.mse_loss(forward_dict['x'][4], forward_dict['xy'][4]),
                     'similarity_loss_y_xy': F.mse_loss(forward_dict['y'][4], forward_dict['xy'][4])}
        loss_dict['total_loss'] = loss_dict['xvae_loss']['loss'] + loss_dict['yvae_loss']['loss'] + \
                                  loss_dict['xyvae_loss']['loss'] + loss_dict['similarity_loss_x_y'] + loss_dict[
                                      'similarity_loss_x_xy'] + loss_dict['similarity_loss_y_xy']
        return loss_dict



import torch
import torch.utils.data
from torch import nn, optim
from torch.nn import functional as F


class StandardVAE(nn.Module):
    def __init__(self, layers_structure, activation_fn, kld_coefficient=1):
        """
        :param layers_structure: An iterable which will form the structure of the encoder, As a result, the structure
        of the decoder will be symmetrical.

        :param activation_fn: A Pointer to the activation function that will be used between the layers.
        """

        super(StandardVAE, self).__init__()
        self.activation_fn = activation_fn
        reversed_layers_structure = layers_structure[::-1]  # the reversed list will form the structure of the decoder.
        self.encoding_layers = nn.ModuleList(
            [nn.Linear(layers_structure[i], layers_structure[i + 1]) for i in
             range(0, len(layers_structure) - 2)])  # Don't create the last layer because it needs to be split.

        # Layers for the mean and the variance of the normal distribution.

        self.mu_layer = nn.Linear(layers_structure[-2], layers_structure[-1])

        self.logvar_layer = nn.Linear(layers_structure[-2], layers_structure[-1])

        self.decoding_layers = nn.ModuleList(
            [nn.Linear(reversed_layers_structure[i], reversed_layers_structure[i + 1]) for i in
             range(0, len(reversed_layers_structure) - 1)])
        self.kld_coefficient = kld_coefficient

    def encode(self, x):
        """
        :param x The input to the network, The function will output the corresponding mu, and logvar.
        """
        for encoding_layer in self.encoding_layers:
            x = self.activation_fn(encoding_layer(x))
        return self.mu_layer(x), self.logvar_layer(x)

    def reparameterize(self, mu, logvar):
        """

        :param mu The first output of the encoding function, represents the mean of the latent distribution
        :param logvar: The second output of the encoding function, will be transformed into the distribution variance.
        :return: The latent representation of the input of the network, using the parameterization trick.
        """
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return mu + (eps * std)

    def decode(self, z):
        """

        :param z: the latent representation of the input which is the output of the reparameterize function.
        :return:
        """
        for layer_number, decoding_layer in enumerate(self.decoding_layers):
            if layer_number < len(self.decoding_layers) - 1:
                z = self.activation_fn(decoding_layer(z))
            else:
                z = decoding_layer(z)
        return z

    def forward(self, input):
        """

        :param input: The input to the first layer.
        :return: List which consists of, the reconstructed input,
        the input itself, the mu and the sigma constructing the distribution from which the output was sampled,
        and the latent representation of the input.
        """
        mu, logvar = self.encode(input)  # self.encode(x.view(-1, 784))
        z = self.reparameterize(mu, logvar)
        return [self.decode(z), input, mu, logvar, z]

    def loss_function(self, *args):
        """
        A function the given the forward array of a sample, calculates the loss. :param args: Forwrad array :return:
        A dictionary which describes ever component of the loss i.e Reconstruction_Loss, KLD loss, and the total loss.
        """
        recons = args[0]
        input = args[1]
        mu = args[2]
        log_var = args[3]
        kld_weight = self.kld_coefficient
        recons_loss = F.mse_loss(recons, input)
        kld_loss = torch.mean(-0.5 * torch.sum(1 + log_var - mu ** 2 - log_var.exp(), dim=1), dim=0)
        loss = recons_loss + kld_weight * kld_loss
        return {'loss': loss, 'Reconstruction_Loss': recons_loss, 'KLD': -kld_loss}



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

VERSION = '0.4.7'
PACKAGE_NAME = 'integration_tools'
AUTHOR = 'Sharon Komissarov'
AUTHOR_EMAIL = 'sharon200102@gmail.com'
URL = 'https://github.com/sharon200102/integration_tools'

LICENSE = ' MIT License'
DESCRIPTION = 'A package in development stages which allows to intgrate few data-sets using machine learning tools'
LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESC_TYPE = "text/markdown"

INSTALL_REQUIRES = [
      'torch',
      'pandas',
      'typing',
      'argparse'
      
]

setup(name=PACKAGE_NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESC_TYPE,
      author=AUTHOR,
      license=LICENSE,
      author_email=AUTHOR_EMAIL,
      url=URL,
      install_requires=INSTALL_REQUIRES,
      packages=find_packages()
      )



import torch
from ...utils.data.data_classes import patient_samples_single_tp


class ToTensor(object):
    """Convert ndarrays in sample to Tensors."""

    def __call__(self, sample):
        sample_dict_copy = dict(sample.patient_dict)
        if sample_dict_copy.get('FIELD0', None) is not None:
            sample_dict_copy['FIELD0'] = torch.tensor(sample.patient_dict['FIELD0'].values.astype('float64')).type(torch.FloatTensor)
        if sample_dict_copy.get('FIELD1', None) is not None:
            sample_dict_copy['FIELD1'] = torch.tensor(sample.patient_dict['FIELD1'].values.astype('float64')).type(torch.FloatTensor)
        return patient_samples_single_tp(sample_dict_copy)

